"""TensorFlow Binary Classification with Text data"""
# Internal modules
from .. import datum
from ..orm import *
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import LabelBinarizer
from sklearn.feature_extraction.text import CountVectorizer


def fn_build(features_shape, label_shape, **hp):
	model = tf.keras.models.Sequential()
	model.add(l.Dense(15, activation='relu', kernel_initializer='he_uniform'))
	model.add(l.Dropout(0.30))
	model.add(l.Dense(15, activation='relu', kernel_initializer='he_uniform'))
	model.add(l.Dropout(0.30))
	model.add(l.Dense(15, activation='relu', kernel_initializer='he_uniform'))
	model.add(l.Dense(units=label_shape[0], activation='sigmoid', kernel_initializer='glorot_uniform'))
	return model


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):
	model.compile(
		loss = loser
		, optimizer = optimizer
		, metrics = ['accuracy']
	)
	model.fit(
		samples_train['features'], samples_train['labels']
		, validation_data = (samples_evaluate['features'], samples_evaluate['labels'])
		, verbose = 0
		, batch_size = 3
		, epochs = 15
		, callbacks = [tf.keras.callbacks.History()]
	)
	return model

def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	hyperparameters = {
		"neuron_count": [25]
		, "epochs": [15]
	}

	file_path = datum.get_path('spam.csv')

	d_id = Dataset.Text.from_path(
		file_path = file_path
		, source_file_format = 'csv'
		, name = 'text test dataset'
		, dtype = None
	).id
	
	label_column = 'label'
	l_id = Label.from_dataset(dataset_id=d_id, columns=[label_column]).id
	f_id = Feature.from_dataset(dataset_id=d_id, exclude_columns=[label_column]).id

	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.18
		size_validation = 0.14

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = l_id
		, size_test = size_test
		, size_validation = size_validation
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(splitset_id=s_id, fold_count=fold_count).id
	else:
		fs_id = None

	LabelCoder.from_label(
		label_id=l_id, sklearn_preprocess=LabelBinarizer(sparse_output=False)
	)

	e_id = Encoderset.from_feature(feature_id=f_id).id
	FeatureCoder.from_encoderset(
		encoderset_id=e_id
		, sklearn_preprocess = CountVectorizer(max_features = 200)
		, columns=['TextData']
	)

	a_id = Algorithm.make(
		library = "keras"
		, analysis_type = "classification_binary"
		, fn_build = fn_build
		, fn_train = fn_train
	).id

	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = h_id
		, repeat_count = repeat_count
		, permute_count = permute_count
	)
	return queue


from pprint import pformat
from operator import ge, le
import tensorflow as tf
from math import ceil
import numpy as np

## --- Default Architectures ---
# `**hp` must be included

# `select_fn_lose()`
def lose_regression(**hp):
	loser = tf.keras.losses.MeanAbsoluteError()
	return loser

def lose_binary(**hp):
	loser = tf.keras.losses.BinaryCrossentropy()
	return loser

def lose_multiclass(**hp):
	loser = tf.keras.losses.CategoricalCrossentropy()
	return loser

# `select_fn_optimize()`
def optimize(**hp):
	optimizer = tf.keras.optimizers.Adamax(learning_rate=0.01)
	return optimizer

# `select_fn_predict()`
def predict_multiclass(model, samples_predict):
	# Shows the probabilities of each class coming out of softmax neurons:
	# array([[9.9990356e-01, 9.6374511e-05, 3.3754202e-10],...])
	probabilities = model.predict(samples_predict['features'])
	# This is the official keras replacement for multiclass `.predict_classes()`
	# Returns one ordinal array per sample: `[[0][1][2][3]]` 
	prediction = np.argmax(probabilities, axis=-1)
	return prediction, probabilities

def predict_binary(model, samples_predict):
	# Sigmoid output is between 0 and 1.
	# It's not technically a probability, but it is still easy to interpret.
	probability = model.predict(samples_predict['features'])
	# This is the official keras replacement for binary classes `.predict_classes()`.
	# Returns one array per sample: `[[0][1][0][1]]`.
	prediction = (probability > 0.5).astype("int32")
	return prediction, probability

def predict_regression(model, samples_predict):
	prediction = model.predict(samples_predict['features'])
	# ^ Output is a single value, not `probability, prediction`
	return prediction


class TrainingCallback():
	class MetricCutoff(tf.keras.callbacks.Callback):
		"""
		- Worried that these inner functions are not pickling during multi-processing.
		https://stackoverflow.com/a/8805244/5739514
		"""
		def __init__(self, thresholds:list):
			"""
			# Tested with keras:2.4.3, tensorflow:2.3.1
			# `thresholds` is list of dictionaries with 1 dict per metric.
			metrics_cuttoffs = [
				{"metric":"val_acc", "cutoff":0.94, "above_or_below":"above"},
				{"metric":"acc", "cutoff":0.90, "above_or_below":"above"},
				{"metric":"val_loss", "cutoff":0.26, "above_or_below":"below"},
				{"metric":"loss", "cutoff":0.30, "above_or_below":"below"},
			]
			# Only stops training early if all user-specified metrics are satisfied.
			# `above_or_below`: where 'above' means `>=` and 'below' means `<=`.
			"""
			self.thresholds = thresholds
			

		def on_epoch_end(self, epoch, logs=None):
			logs = logs or {}
			# Check each user-defined threshold to see if it is satisfied.
			for threshold in self.thresholds:
				metric = logs.get(threshold['metric'])
				if (metric is None):
					raise Exception(dedent(f"""
					Yikes - The metric named '{threshold['metric']}' not found when running `logs.get('{threshold['metric']}')`
					during `TrainingCallback.MetricCutoff.on_epoch_end`.
					"""))
				cutoff = threshold['cutoff']

				above_or_below = threshold['above_or_below']
				if (above_or_below == 'above'):
					statement = ge(metric, cutoff)
				elif (above_or_below == 'below'):
					statement = le(metric, cutoff)
				else:
					raise Exception(dedent(f"""
					Yikes - Value for key 'above_or_below' must be either string 'above' or 'below'.
					You provided:{above_or_below}
					"""))

				if (statement == False):
					break # Out of for loop.
					
			if (statement == False):
				pass # Thresholds not satisfied, so move on to the next epoch.
			elif (statement == True):
				# However, if the for loop actually finishes, then all metrics are satisfied.
				print(
					f":: Epoch #{epoch} ::\n" \
					f"Congratulations - satisfied early stopping thresholds defined in `MetricCutoff` callback:\n"\
					f"{pformat(self.thresholds)}\n"
				)
				self.model.stop_training = True


## --- Batching ---
# Used for a manual training loop without Keras.
def batcher(features:object, labels:object, batch_size:int=5):
	"""
	- `np.array_split` allows for subarrays to be of different sizes, which is rare.
	  https://numpy.org/doc/stable/reference/generated/numpy.array_split.html 
	- If there is a remainder, it will evenly distribute samples into the other arrays.
	- Have not tested this with >= 3D data yet.
	"""
	rows_per_batch = ceil(features.shape[0]/batch_size)

	batched_features = np.array_split(features, rows_per_batch)
	batched_features = np.array(batched_features, dtype=object)

	batched_labels = np.array_split(labels, rows_per_batch)
	batched_labels = np.array(batched_labels, dtype=object)
	return batched_features, batched_labels



"""TensorFlow Forecasting with Image data"""
# Internal modules
from ..orm import *
from ..utils.encoding import div255, mult255
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import FunctionTransformer


def fn_build(features_shape, label_shape, **hp):
	"""
	- Model size was too large for SQLite blob... need to rework to store it on the FS.
	- incoming features_shape = frame* channels * rows * columns
	- https://keras.io/api/layers/reshaping_layers/reshape/
	- Model size was too large for SQLite blob... need to rework to store it on the FS.
	- ConvLSTM1D is still in nightly build so use ConvLSTM2D
	- https://www.tensorflow.org/api_docs/python/tf/keras/layers/ConvLSTM2D
	- If data_format='channels_last' 5D tensor with shape: (samples, time, rows, cols, channels)
	"""
	m = tf.keras.models.Sequential()
	m.add(l.Conv1D(64*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.MaxPool1D( 2, padding='same'))
	m.add(l.Conv1D(32*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.MaxPool1D( 2, padding='same'))
	m.add(l.Conv1D(16*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.MaxPool1D( 2, padding='same'))

	# decoding architecture
	m.add(l.Conv1D(16*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.UpSampling1D(2))
	m.add(l.Conv1D(32*hp['multiplier'], 3, activation=hp['activation'], padding='same'))
	m.add(l.UpSampling1D(2))
	m.add(l.Conv1D(64*hp['multiplier'], 3, activation=hp['activation']))
	m.add(l.UpSampling1D(2))
	m.add(l.Conv1D(50, 3, activation='relu', padding='same'))# removing sigmoid
	return m


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):
	model.compile(
		optimizer=optimizer
		, loss=loser
		, metrics=['mean_squared_error']
	)
	
	model.fit(
		samples_train["features"]
		, samples_train["labels"]
		, validation_data = (
			samples_evaluate["features"]
			, samples_evaluate["labels"]
		)
		, verbose = 0
		, batch_size = hp['batch_size']
		, callbacks=[tf.keras.callbacks.History()]
		, epochs = hp['epoch_count']
	)
	return model



def make_queue(repeat_count:int=1, fold_count:int=None):
	folder_path = 'remote_datum/image/liberty_moon/images'
	di_id = Dataset.Image.from_folder_pillow(
		folder_path=folder_path, ingest=False, dtype='float64'
	).id

	f_id = Feature.from_dataset(dataset_id=di_id).id
	Window.from_feature(feature_id=f_id, size_window=1, size_shift=2)
	e_id = Encoderset.from_feature(feature_id=f_id).id
	FeatureCoder.from_encoderset(
		encoderset_id = e_id
		, sklearn_preprocess = FunctionTransformer(div255, inverse_func=mult255)
		, dtypes = 'float64'
	)
	FeatureShaper.from_feature(feature_id=f_id, reshape_indices=(0,3,4))

	if (fold_count is not None):
		size_test = 0.15
		size_validation = None
	elif (fold_count is None):
		size_test = 0.15
		size_validation = None#small dataset

	s_id = Splitset.make(
		feature_ids = [f_id]
		, size_test = size_test
		, size_validation = size_validation
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(
			splitset_id=s_id, fold_count=fold_count
		).id
	else:
		fs_id = None

	a_id = Algorithm.make(
		library = "keras"
		, analysis_type = "regression"
		, fn_build = fn_build
		, fn_train = fn_train
		# , fn_lose = keras_image_forecast_fn_lose
	).id

	hyperparameters = dict(
		epoch_count = [12]
		, batch_size = [3]
		, cnn_init = ['he_normal']
		, activation = ['relu']
		, multiplier = [3]
	)
	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = h_id
		, repeat_count = repeat_count
	)
	return queue



"""Sets defaults when assembling architecture in Queue object"""
from . import tensorflow, pytorch
from textwrap import dedent


def select_fn_lose(
	library:str,
	analysis_type:str
):      
	fn_lose = None
	if (library == 'keras'):
		if (analysis_type == 'regression'):
			fn_lose = tensorflow.lose_regression
		elif (analysis_type == 'classification_binary'):
			fn_lose = tensorflow.lose_binary
		elif (analysis_type == 'classification_multi'):
			fn_lose = tensorflow.lose_multiclass
	elif (library == 'pytorch'):
		if (analysis_type == 'regression'):
			fn_lose = pytorch.lose_regression
		elif (analysis_type == 'classification_binary'):
			fn_lose = pytorch.lose_binary
		elif (analysis_type == 'classification_multi'):
			fn_lose = pytorch.lose_multiclass
	# After each of the predefined approaches above, check if it is still undefined.
	if fn_lose is None:
		raise Exception(dedent("""
		Yikes - You did not provide a `fn_lose`,
		and we don't have an automated function for your combination of 'library' and 'analysis_type'
		"""))
	return fn_lose


def select_fn_optimize(library:str):
	fn_optimize = None
	if (library == 'keras'):
		fn_optimize = tensorflow.optimize
	elif (library == 'pytorch'):
		fn_optimize = pytorch.optimize
	# After each of the predefined approaches above, check if it is still undefined.
	if (fn_optimize is None):
		raise Exception(dedent("""
		Yikes - You did not provide a `fn_optimize`,
		and we don't have an automated function for your 'library'
		"""))
	return fn_optimize


def select_fn_predict(
	library:str,
	analysis_type:str
):
	fn_predict = None
	if (library == 'keras'):
		if (analysis_type == 'classification_multi'):
			fn_predict = tensorflow.predict_multiclass
		elif (analysis_type == 'classification_binary'):
			fn_predict = tensorflow.predict_binary
		elif (analysis_type == 'regression'):
			fn_predict = tensorflow.predict_regression
	elif (library == 'pytorch'):
		if (analysis_type == 'classification_multi'):
			fn_predict = pytorch.predict_multiclass
		elif (analysis_type == 'classification_binary'):
			fn_predict = pytorch.predict_binary
		elif (analysis_type == 'regression'):
			fn_predict = pytorch.predict_regression

	# After each of the predefined approaches above, check if it is still undefined.
	if fn_predict is None:
		raise Exception(dedent("""
		Yikes - You did not provide a `fn_predict`,
		and we don't have an automated function for your combination of 'library' and 'analysis_type'
		"""))
	return fn_predict



"""TensorFlow Binary Classification with Sequence data"""
# Internal modules
from .. import datum
from ..orm import *
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import StandardScaler
import numpy as np


def fn_build(features_shape, label_shape, **hp):    
	m = tf.keras.models.Sequential()
	m.add(l.LSTM(
		hp['neuron_count']
		, input_shape=(features_shape[0], features_shape[1])
	))
	m.add(l.Dense(units=label_shape[0], activation='sigmoid'))
	return m


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):
	model.compile(
		loss=loser
		, optimizer=optimizer
		, metrics=['accuracy']
	)
	model.fit(
		samples_train['features'], samples_train['labels']
		, validation_data = (samples_evaluate['features'], samples_evaluate['labels'])
		, verbose = 0
		, batch_size = hp['batch_size']
		, epochs = hp['epochs']
		, callbacks = [tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	df = datum.to_pandas('epilepsy.parquet')
	# testing Featurepolater 3D.
	df['sensor_1'][999] = np.NaN
	df['sensor_1'][0] = np.NaN
	df['sensor_150'][130] = np.NaN
	df['sensor_152'][22] = np.NaN
	df['sensor_170'][0] = np.NaN
	
	label_df = df[['seizure']]
	dt_id = Dataset.Tabular.from_pandas(label_df).id
	l_id = Label.from_dataset(dataset_id=dt_id, columns='seizure').id

	sensor_arr3D = df.drop(columns=['seizure']).to_numpy().reshape(1000,178,1).astype('float64')	
	ds_id = Dataset.Sequence.from_numpy(sensor_arr3D).id
	f_id = Feature.from_dataset(dataset_id=ds_id).id
	
	i_id = Interpolaterset.from_feature(feature_id=f_id).id
	FeatureInterpolater.from_interpolaterset(interpolaterset_id=i_id, dtypes="float64")
	
	e_id = Encoderset.from_feature(feature_id=f_id).id
	FeatureCoder.from_encoderset(
		encoderset_id = e_id
		, sklearn_preprocess = StandardScaler()
		, columns = ['0']
	)
	
	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.22
		size_validation = 0.12

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = l_id
		, size_test = size_test
		, size_validation = size_validation
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(
			splitset_id=s_id, fold_count=fold_count
		).id
	else:
		fs_id = None
	
	a_id = Algorithm.make(
		library = "keras"
		, analysis_type = "classification_binary"
		, fn_build = fn_build
		, fn_train = fn_train
	).id
	
	hyperparameters = {
		"neuron_count": [25]
		, "batch_size": [8]
		, "epochs": [5]
	}
	
	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, hyperparamset_id = h_id
		, foldset_id = fs_id
		, repeat_count = repeat_count
		, permute_count = permute_count
	)
	return queue



"""
Installation
---
Create `/aiqc` app_dirs (an os-agnostic folder).
Create `config.json` for storing settings.
Create `aiqc.sqlite3` database.
"""
# Python modules
import os, sys, platform, json, importlib, datetime
# External modules
from appdirs import user_data_dir


app_dir_no_trailing_slash = user_data_dir("aiqc")
# Adds either a trailing slash or backslashes depending on OS.
app_dir = os.path.join(app_dir_no_trailing_slash, '')
default_config_path = app_dir + "config.json"
default_db_path = app_dir + "aiqc.sqlite3"
# API URL format 'https://{api_id}.execute-api.{region}.amazonaws.com/{stage}/'.
# aws_api_root = "https://qzdvq719pa.execute-api.us-east-1.amazonaws.com/Stage_AIQC/"


#==================================================
# FOLDER
#==================================================
def check_exists_folder():
	# If Windows does not have permission to read the folder, it will fail when trailing backslashes \\ provided.
	app_dir_exists = os.path.exists(app_dir_no_trailing_slash)
	if app_dir_exists:
		print(f"\n=> Success - the following file path already exists on your system:\n{app_dir}\n")
		return True
	else:
		print(
			f"=> Info - it appears the following folder does not exist on your system:\n{app_dir}\n\n" \
			f"=> Fix - you can attempt to fix this by running `aiqc.config.create_folder()`.\n"
		)
		return False


def create_folder():
	app_dir_exists = check_exists_folder()
	if (app_dir_exists):
		print(f"\n=> Info - skipping folder creation as folder already exists at file path:\n{app_dir}\n")
	else:
		try:
			"""
			- `makedirs` will create any missing intermediary dir(s) in addition to the target dir.
			- Whereas `mkdir` only creates the target dir and fails if intermediary dir(s) are missing.
			- If this break for whatever reason, could also try out `path.mkdir(parents=True)`.
			"""
			os.makedirs(app_dir)
			# if os.name == 'nt':
			#   # Windows: backslashes \ and double backslashes \\
			#   command = 'mkdir ' + app_dir
			#   os.system(command)
			# else:
			#   # posix (mac and linux)
			#   command = 'mkdir -p "' + app_dir + '"'
			#   os.system(command)
		except:
			raise OSError(f"\n=> Yikes - Local system failed to execute:\n`os.makedirs('{app_dir}')\n")
		print(
			f"=> Success - created folder at file path:\n{app_dir}\n\n" \
			f"=> Next run `aiqc.config.create_config()`.\n"
		)


def check_permissions_folder():
	app_dir_exists = check_exists_folder()
	if (app_dir_exists):
		# Windows `os.access()` always returning True even when I have verify permissions are in fact denied.
		if (os.name == 'nt'):
			# Test write.
			file_name = "aiqc_test_permissions.txt"
			
			def permissions_fail_info():
				# We don't want an error here because it needs to return False.
				print(
					f"=> Yikes - your operating system user does not have permission to write to file path:\n{app_dir}\n\n" \
					f"=> Fix - you can attempt to fix this by running `aiqc.config.grant_permissions_folder()`.\n"
				)

			try:
				cmd_file_create = 'echo "test" >> ' + app_dir + file_name
				write_response = os.system(cmd_file_create)
			except:
				permissions_fail_info()
				return False

			if (write_response != 0):
				permissions_fail_info()
				return False
			else:
				# Test read.
				try:
					read_response = os.system("type " + app_dir + file_name)
				except:
					permissions_fail_info()
					return False

				if (read_response != 0):
					permissions_fail_info()
					return False
				else:
					cmd_file_delete = "erase " + app_dir + file_name
					os.system(cmd_file_delete)
					print(f"\n=> Success - your operating system user can read from and write to file path:\n{app_dir}\n")
					return True

		else:
			# posix
			# https://www.geeksforgeeks.org/python-os-access-method/
			readable = os.access(app_dir, os.R_OK)
			writeable = os.access(app_dir, os.W_OK)

			if (readable and writeable):
				print(f"\n=> Success - your operating system user can read from and write to file path:\n{app_dir}\n")
				return True
			else:
				if not readable:
					print(f"\n=> Yikes - your operating system user does not have permission to read from file path:\n{app_dir}\n")
				if not writeable:
					print(f"\n=> Yikes - your operating system user does not have permission to write to file path:\n{app_dir}\n")
				if not readable or not writeable:
					print("\n=> Fix - you can attempt to fix this by running `aiqc.config.grant_permissions_folder()`.\n")
					return False
	else:
		return False


def grant_permissions_folder():
	permissions = check_permissions_folder()
	if (permissions):
		print(f"\n=> Info - skipping as you already have permissions to read from and write to file path:\n{app_dir}\n")
	else:
		try:
			if (os.name == 'nt'):
				# Windows ICACLS permissions: https://www.educative.io/edpresso/what-is-chmod-in-windows
				# Works in Windows Command Prompt and `os.system()`, but not PowerShell.
				# Does not work with trailing backslashes \\
				command = 'icacls "' + app_dir_no_trailing_slash + '" /grant users:(F) /c'
				os.system(command)
			elif (os.name != 'nt'):
				# posix
				command = 'chmod +wr ' + '"' + app_dir + '"'
				os.system(command)
		except:
			print(
				f"=> Yikes - error failed to execute this system command:\n{command}\n\n" \
				f"===================================\n"
			)
			raise
		
		permissions = check_permissions_folder()
		if permissions:
			print(f"\n=> Success - granted system permissions to read and write from file path:\n{app_dir}\n")
		else:
			print(f"\n=> Yikes - failed to grant system permissions to read and write from file path:\n{app_dir}\n")

#==================================================
# CONFIG
#==================================================

def get_config():
	aiqc_config_exists = os.path.exists(default_config_path)
	if aiqc_config_exists:
		with open(default_config_path, 'r') as aiqc_config_file:
			aiqc_config = json.load(aiqc_config_file)
			return aiqc_config
	else: 
		print("\n=> Welcome to AIQC.\nTo get started, run `aiqc.setup()`.\n")


def create_config():
	#check if folder exists
	folder_exists = check_exists_folder()
	if folder_exists:
		config_exists = os.path.exists(default_config_path)
		if not config_exists:
			aiqc_config = {
				"created_at": str(datetime.datetime.now())
				, "config_path": default_config_path
				, "db_path": default_db_path
				, "sys.version": sys.version
				, "platform.python_implementation()": platform.python_implementation()
				, "sys.prefix": sys.prefix
				, "os.name": os.name
				, "platform.version()": platform.version()
				, "platform.java_ver()": platform.java_ver()
				, "platform.win32_ver()": platform.win32_ver()
				, "platform.libc_ver()": platform.libc_ver()
				, "platform.mac_ver()": platform.mac_ver()
				# , "aws_api_key": None
			}
			
			try:
				with open(default_config_path, 'w') as aiqc_config_file:
					json.dump(aiqc_config, aiqc_config_file)
			except:
				print(
					f"=> Yikes - failed to create config file at path:\n{default_config_path}\n\n" \
					f"=> Fix - you can attempt to fix this by running `aiqc.config.check_permissions_folder()`.\n" \
					f"==================================="
				)
				raise
			print(f"\n=> Success - created config file for settings at path:\n{default_config_path}\n")
			importlib.reload(sys.modules[__name__])
		else:
			print(f"\n=> Info - skipping as config file already exists at path:\n{default_config_path}\n")
	print("\n=> Next run `aiqc.orm.create_db()`.\n")


def delete_config(confirm:bool=False):
	aiqc_config = get_config()
	if aiqc_config is None:
		print("\n=> Info - skipping as there is no config file to delete.\n")
	else:
		if confirm:
			config_path = aiqc_config['config_path']
			try:
				os.remove(config_path)
			except:
				print(
					f"=> Yikes - failed to delete config file at path:\n{config_path}\n\n" \
					f"===================================\n" \
				)
				raise
			print(f"\n=> Success - deleted config file at path:\n{config_path}\n")
			importlib.reload(sys.modules[__name__])   
		else:
			print("\n=> Info - skipping deletion because `confirm` arg not set to boolean `True`.\n")


def update_config(kv:dict):
	aiqc_config = get_config()
	if aiqc_config is None:
		print("\n=> Info - there is no config file to update.\n")
	else:
		for k, v in kv.items():
			aiqc_config[k] = v      
		config_path = aiqc_config['config_path']
		
		try:
			with open(config_path, 'w') as aiqc_config_file:
				json.dump(aiqc_config, aiqc_config_file)
		except:
			print(
				f"=> Yikes - failed to update config file at path:\n{config_path}\n\n" \
				f"===================================\n"
			)
			raise
		print(f"\n=> Success - updated configuration settings at path:\n{config_path}\n")
		importlib.reload(sys.modules[__name__])


# def config_add_aws_api_key(api_key:str):
# 	config_test_aws_api_key(api_key)
# 	update_config(kv=dict(aws_api_key=api_key))


# def config_test_aws_api_key(api_key:str):
# 	print(f"\n=> Info - testing API key via connection to AIQC AWS API Gateway:\n{aws_api_root}\n")
# 	response = requests.get(
# 		url=aws_api_root, headers={"x-api-key":api_key}
# 	)
# 	status_code = response.status_code
# 	if (status_code==200):
# 		print("\n=> Success - was able to connect to root of AIQC AWS API Gateway.\n")
# 	else:
# 		raise Exception(f"\n=> Yikes - failed to connect to root of AIQC AWS API Gateway.\nHTTP Error Code = {status_code}.\n")


# def _aws_get_api_key():
# 	api_key = get_config()['aws_api_key']
# 	if (api_key is None):
# 		raise Exception("\nYikes - `config['aws_api_key']` has not been set.\nRun `config_add_aws_api_key(api_key)` to define it.\n")
# 	elif (api_key is not None):
# 		return api_key



"""TensorFlow Forecasting with Tabular data"""
# Internal modules
from .. import datum
from ..orm import *
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l
from sklearn.preprocessing import StandardScaler, RobustScaler
import numpy as np


def fn_build(features_shape, label_shape, **hp):
	m = tf.keras.models.Sequential()
	m(l.GRU(
			hp['neuron_count']
			, input_shape=(features_shape[0], features_shape[1])
			, return_sequences=False
			, activation='tanh'
	))
	# Automatically flattens.
	m(l.Dense(label_shape[0]*label_shape[1]*hp['dense_multiplier'], activation='tanh'))
	m(l.Dropout(0.3))
	m(l.Dense(label_shape[0]*label_shape[1], activation='tanh'))
	m(l.Dropout(0.3))
	# Reshape to be 3D.
	m(l.Reshape((label_shape[0], label_shape[1])))
	
	return m


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):
	model.compile(
		loss=loser
		, optimizer=optimizer
		, metrics=['mean_squared_error']
	)
		
	model.fit(
		samples_train['features'], samples_train['features']
		, validation_data = (samples_evaluate['features'], samples_evaluate['features'])
		, verbose = 0
		, batch_size = hp['batch_size']
		, epochs = hp['epochs']
		, callbacks = [tf.keras.callbacks.History()]
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None, permute_count:int=3):
	df = datum.to_pandas('delhi_climate.parquet')
	df['temperature'][0] = np.NaN
	df['temperature'][13] = np.NaN
	
	d_id = Dataset.Tabular.from_pandas(dataframe=df).id

	f_id = Feature.from_dataset(dataset_id=d_id).id
	
	i_id = Interpolaterset.from_feature(feature_id=f_id).id
	FeatureInterpolater.from_interpolaterset(interpolaterset_id=i_id, dtypes=['float64'])

	Window.from_feature(feature_id=f_id, size_window=28, size_shift=14)

	e_id = Encoderset.from_feature(feature_id=f_id).id
	FeatureCoder.from_encoderset(
		encoderset_id = e_id
		, sklearn_preprocess = RobustScaler(copy=False)
		, columns = ['wind', 'pressure']
	)
	FeatureCoder.from_encoderset(
		encoderset_id = e_id
		, sklearn_preprocess = StandardScaler()
		, dtypes = ['float64', 'int64']
	)

	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.17
		size_validation = 0.16

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = None
		, size_test = size_test
		, size_validation = size_validation
		, bin_count = None
		, unsupervised_stratify_col = 'day_of_year'
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(
			splitset_id=s_id, fold_count=fold_count
		).id
	else:
		fs_id = None

	a_id = Algorithm.make(
		library = "keras"
		, analysis_type = "regression"
		, fn_build = fn_build
		, fn_train = fn_train
	).id

	hyperparameters = {
		"neuron_count": [8]
		, "batch_size": [8]
		, "epochs": [12]
		, "dense_multiplier": [1]
	}

	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = h_id
		, repeat_count = repeat_count
		, permute_count = permute_count
	)
	return queue



from .config import create_folder, create_config
from .utils.wrangle import listify
from .orm import *


name = "aiqc"


def setup():
	create_folder()
	create_config()
	# comes from orm
	create_db()


"""
High-Level API
├── Documentation = https://aiqc.readthedocs.io/en/latest/notebooks/api_high_level.html
└── Examples = https://aiqc.readthedocs.io/en/latest/tutorials.html

These classes really just bundle a bunch of low-level API commands they don't instantiate 
real `Pipeline` objects bur rather `splitsets`. They use `__new__` because `__init__` 
cannot return anything.
"""
class Pipeline:
	def parse_tabular_input(dataFrame_or_filePath:object, dtype:object=None):
		"""Create the dataset from either df or file."""
		d = dataFrame_or_filePath
		data_type = str(type(d))
		if (data_type == "<class 'pandas.core.frame.DataFrame'>"):
			dataset = Dataset.Tabular.from_pandas(dataframe=d, dtype=dtype)
		elif (data_type == "<class 'str'>"):
			if ('.csv' in d):
				source_file_format='csv'
			elif ('.tsv' in d):
				source_file_format='tsv'
			elif ('.parquet' in d):
				source_file_format='parquet'
			else:
				raise Exception(dedent("""
				Yikes - None of the following file extensions were found in the path you provided:
				'.csv', '.tsv', '.parquet'
				"""))
			dataset = Dataset.Tabular.from_path(
				file_path = d
				, source_file_format = source_file_format
				, dtype = dtype
			)
		else:
			raise Exception("\nYikes - The `dataFrame_or_filePath` is neither a string nor a Pandas dataframe.\n")
		return dataset


	class Tabular:
		def __new__(
			cls
			, df_or_path:object
			, dtype:object = None
			
			, feature_cols_excluded:list = None
			, feature_interpolaters:list = None
			, feature_window:dict = None
			, feature_encoders:list = None
			, feature_reshape_indices:tuple = None

			, label_column:str = None
			, label_interpolater:dict = None
			, label_encoder:dict = None

			, size_test:float = None
			, size_validation:float = None
			, fold_count:int = None
			, bin_count:int = None
		):
			feature_cols_excluded = listify(feature_cols_excluded)
			feature_interpolaters = listify(feature_interpolaters)
			feature_encoders = listify(feature_encoders)
			label_column = listify(label_column)

			d_id = Pipeline.parse_tabular_input(
				dataFrame_or_filePath=df_or_path, dtype=dtype
			).id

			if (label_column is not None):
				l_id = Label.from_dataset(dataset_id=d_id, columns=label_column).id
				if (label_interpolater is not None):
					LabelInterpolater.from_label(label_id=l_id, **label_interpolater)
				if (label_encoder is not None): 
					LabelCoder.from_label(label_id=l_id, **label_encoder)
			elif (label_column is None):
				# Needs to know if label exists so that it can exlcude it.
				l_id = None

			if (feature_cols_excluded is None):
				if (label_column is not None):
					f_id = Feature.from_dataset(dataset_id=d_id, exclude_columns=label_column).id
				# Unsupervised.
				elif (label_column is None):
					f_id = Feature.from_dataset(dataset_id=d_id).id
			elif (feature_cols_excluded is not None):
				f_id = Feature.from_dataset(dataset_id=d_id, exclude_columns=feature_cols_excluded).id

			if (feature_interpolaters is not None):
				i_id = Interpolaterset.from_feature(feature_id=f_id).id
				for fp in feature_interpolaters:
					FeatureInterpolater.from_interpolaterset(i_id, **fp)

			if (feature_window is not None):
				Window.from_feature(feature_id=f_id, **feature_window)

			if (feature_encoders is not None):					
				e_id = Encoderset.from_feature(feature_id=f_id).id
				for fc in feature_encoders:
					FeatureCoder.from_encoderset(encoderset_id=e_id, **fc)

			if (feature_reshape_indices is not None):
				FeatureShaper.from_feature(feature_id=f_id, reshape_indices=feature_reshape_indices)

			splitset = Splitset.make(
				feature_ids = [f_id]
				, label_id = l_id
				, size_test = size_test
				, size_validation = size_validation
				, bin_count = bin_count
			)
			
			if (fold_count is not None):
				Foldset.from_splitset(
					splitset_id=splitset.id, fold_count=fold_count, bin_count=bin_count
				)
			return splitset


	class Sequence:
		def __new__(
			cls
			, feature_ndarray3D_or_npyPath:object
			, feature_dtype:object = None
			, feature_cols_excluded:list = None
			, feature_interpolaters:list = None
			, feature_window:dict = None
			, feature_encoders:list = None
			, feature_reshape_indices:tuple = None
			
			, label_df_or_path:object = None
			, label_dtype:object = None
			, label_column:str = None
			, label_interpolater:dict = None
			, label_encoder:dict = None
			
			, size_test:float = None
			, size_validation:float = None
			, fold_count:int = None
			, bin_count:int = None
		):
			feature_cols_excluded = listify(feature_cols_excluded)
			feature_interpolaters = listify(feature_interpolaters)
			feature_encoders = listify(feature_encoders)
			label_column = listify(label_column)

			if (
				((label_df_or_path is None) and (label_column is not None))
				or
				((label_df_or_path is not None) and (label_column is None))
			):
				raise Exception("\nYikes - `label_df_or_path` and `label_column` are either used together or not at all.\n")

			# ------ SEQUENCE FEATURE ------
			ds_id = Dataset.Sequence.from_numpy(
				ndarray3D_or_npyPath=feature_ndarray3D_or_npyPath,
				dtype=feature_dtype
			).id

			if (feature_cols_excluded is not None):
				f_id = Feature.from_dataset(dataset_id=ds_id, exclude_columns=feature_cols_excluded).id
			elif (feature_cols_excluded is None):
				f_id = Feature.from_dataset(dataset_id=ds_id).id

			if (feature_interpolaters is not None):
				i_id = Interpolaterset.from_feature(feature_id=f_id).id
				for fp in feature_interpolaters:
					FeatureInterpolater.from_interpolaterset(i_id, **fp)					

			if (feature_window is not None):
				Window.from_feature(feature_id=f_id, **feature_window)

			if (feature_encoders is not None):					
				e_id = Encoderset.from_feature(feature_id=f_id).id
				for fc in feature_encoders:
					FeatureCoder.from_encoderset(encoderset_id=e_id, **fc)

			if (feature_reshape_indices is not None):
				FeatureShaper.from_feature(feature_id=f_id, reshape_indices=feature_reshape_indices)

			# ------ TABULAR LABEL ------
			if (label_df_or_path is not None):
				dt_id = Pipeline.parse_tabular_input(
					dataFrame_or_filePath = label_df_or_path
					, dtype = label_dtype
				).id
				# Tabular-based Label.
				l_id = Label.from_dataset(dataset_id=dt_id, columns=label_column).id
				if (label_interpolater is not None):
					LabelInterpolater.from_label(label_id=l_id, **label_interpolater)					
				if (label_encoder is not None): 
					LabelCoder.from_label(label_id=l_id, **label_encoder)
			elif (label_df_or_path is None):
				l_id = None

			splitset = Splitset.make(
				feature_ids = [f_id]
				, label_id = l_id
				, size_test = size_test
				, size_validation = size_validation
				, bin_count = bin_count
			)

			if (fold_count is not None):
				Foldset.from_splitset(
					splitset_id=splitset.id, fold_count=fold_count, bin_count=bin_count
				)
			return splitset


	class Image:
		def __new__(
			cls
			, feature_folder_or_urls:str
			, feature_dtype:str = None
			, feature_interpolaters:list = None
			, feature_window:dict = None
			, feature_encoders:list = None
			, feature_reshape_indices:tuple = None

			, label_df_or_path:object = None
			, label_dtype:object = None
			, label_column:str = None
			, label_interpolater:dict = None
			, label_encoder:dict = None

			, size_test:float = None
			, size_validation:float = None
			, fold_count:int = None
			, bin_count:int = None
		):
			label_column = listify(label_column)
			feature_interpolaters = listify(feature_interpolaters)
			feature_encoders = listify(feature_encoders)

			if (
				((label_df_or_path is None) and (label_column is not None))
				or
				((label_df_or_path is not None) and (label_column is None))
			):
				raise Exception("\nYikes - `label_df_or_path` and `label_column` are either used together or not at all.\n")

			if (isinstance(feature_folder_or_urls, str)):
				di_id = Dataset.Image.from_folder_pillow(
					folder_path = feature_folder_or_urls
					, dtype = feature_dtype
				).id
			elif (isinstance(feature_folder_or_urls, list)):
				feature_folder_or_urls = listify(feature_folder_or_urls)
				di_id = Dataset.Image.from_urls_pillow(
					urls = feature_folder_or_urls
					, dtype = feature_dtype
				).id
			# Image-based Feature.
			f_id = Feature.from_dataset(dataset_id=di_id).id

			if (feature_interpolaters is not None):
				i_id = Interpolaterset.from_feature(feature_id=f_id).id
				for fp in feature_interpolaters:
					FeatureInterpolater.from_interpolaterset(interpolaterset_id=i_id, **fp)					

			if (feature_window is not None):
				Window.from_feature(feature_id=f_id, **feature_window)

			if (feature_encoders is not None):
				e_id = Encoderset.from_feature(feature_id=f_id).id
				for fc in feature_encoders:
					FeatureCoder.from_encoderset(encoderset_id=e_id, **fc)

			if (feature_reshape_indices is not None):
				FeatureShaper.from_feature(feature_id=f_id, reshape_indices=feature_reshape_indices)

			# # Tabular-based Label.
			if (label_df_or_path is not None):
				dt_id = Pipeline.parse_tabular_input(
					dataFrame_or_filePath = label_df_or_path
					, dtype = label_dtype
				).id
				
				l_id = Label.from_dataset(dataset_id=dt_id, columns=label_column).id
				if (label_interpolater is not None):
					LabelInterpolater.from_label(label_id=l_id, **label_interpolater)
				if (label_encoder is not None): 
					LabelCoder.from_label(label_id=l_id, **label_encoder)

			elif (label_df_or_path is None):
				l_id = None
			
			splitset = Splitset.make(
				feature_ids = [f_id]
				, label_id = l_id
				, size_test = size_test
				, size_validation = size_validation
				, bin_count = bin_count
			)

			if (fold_count is not None):
				Foldset.from_splitset(
					splitset_id=splitset.id, fold_count=fold_count, bin_count=bin_count
				)
			return splitset


class Experiment:
	"""
	- Create Algorithm, Hyperparamset, preprocess, and Queue.
	- Includes `preprocess` because it's weird to encode labels before you know what your final training layer looks like.
	  Also, it's optional, so you'd have to access it from splitset before passing it in.
	- The only pre-existing things that need to be passed in are `splitset_id` and the optional `foldset_id`.

	`encoder_feature`: List of dictionaries describing each encoder to run along with filters for different feature columns.
	`encoder_label`: Single instantiation of an sklearn encoder: e.g. `OneHotEncoder()` that gets applied to the full label array.
	"""
	def __new__(
		cls
		, library:str
		, analysis_type:str
		, fn_build:object
		, fn_train:object
		, splitset_id:int
		, repeat_count:int = 1
		, permute_count:int = 3
		, hide_test:bool = False
		, fn_optimize:object = None
		, fn_predict:object = None
		, fn_lose:object = None
		, hyperparameters:dict = None
		, search_count = None
		, search_percent = None
		, foldset_id:int = None
	):
		a_id = Algorithm.make(
			library = library
			, analysis_type = analysis_type
			, fn_build = fn_build
			, fn_train = fn_train
			, fn_optimize = fn_optimize
			, fn_predict = fn_predict
			, fn_lose = fn_lose
		).id

		if (hyperparameters is not None):
			h_id = Hyperparamset.from_algorithm(
				algorithm_id = a_id
				, hyperparameters = hyperparameters
				, search_count = search_count
				, search_percent = search_percent
			).id
		elif (hyperparameters is None):
			h_id = None

		queue = Queue.from_algorithm(
			algorithm_id = a_id
			, splitset_id = splitset_id
			, repeat_count = repeat_count
			, permute_count = permute_count
			, hide_test = hide_test
			, hyperparamset_id = h_id
			, foldset_id = foldset_id
		)
		return queue



"""TensorFlow Binary Classification with Image data"""
# Internal modules
from .. import datum
from ..utils.tensorflow import TrainingCallback
from ..orm import *
# External modules
import tensorflow as tf
import tensorflow.keras.layers as l


def fn_build(features_shape, label_shape, **hp):
	m = tf.keras.models.Sequential()
	# incoming features_shape = channels * rows * columns
	# https://keras.io/api/layers/reshaping_layers/reshape/
	# https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv1D
	# Conv1D shape = `batch_shape + (steps, input_dim)`
	m.add(l.Reshape(
		(features_shape[1],features_shape[2])#,features_shape[0])#dropping
		, input_shape=features_shape)
	)

	m.add(l.Conv1D(128*hp['neuron_multiply'], kernel_size=hp['kernel_size'], input_shape=features_shape, padding='same', activation='relu', kernel_initializer=hp['cnn_init']))
	m.add(l.MaxPooling1D(pool_size=hp['pool_size']))
	m.add(l.Dropout(hp['dropout']))
	
	m.add(l.Conv1D(256*hp['neuron_multiply'], kernel_size=hp['kernel_size'], padding='same', activation='relu', kernel_initializer=hp['cnn_init']))
	m.add(l.MaxPooling1D(pool_size=hp['pool_size']))
	m.add(l.Dropout(hp['dropout']))

	m.add(l.Flatten())
	m.add(l.Dense(hp['dense_neurons']*hp['neuron_multiply'], activation='relu'))
	m.add(l.Dropout(0.2))
	if hp['include_2nd_dense'] == True:
		m.add(l.Dense(hp['2nd_dense_neurons'], activation='relu'))

	m.add(l.Dense(units=label_shape[0], activation='sigmoid'))
	return m


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):   
	model.compile(
		optimizer=optimizer
		, loss=loser
		, metrics=['accuracy']
	)

	metrics_cuttoffs = [
		{"metric":"val_accuracy", "cutoff":0.70, "above_or_below":"above"},
		{"metric":"accuracy", "cutoff":0.70, "above_or_below":"above"},
		{"metric":"val_loss", "cutoff":0.50, "above_or_below":"below"},
		{"metric":"loss", "cutoff":0.50, "above_or_below":"below"}
	]
	cutoffs = TrainingCallback.MetricCutoff(metrics_cuttoffs)
	
	model.fit(
		samples_train["features"]
		, samples_train["labels"]
		, validation_data = (
			samples_evaluate["features"]
			, samples_evaluate["labels"]
		)
		, verbose = 0
		, batch_size = hp['batch_size']
		, callbacks=[tf.keras.callbacks.History(), cutoffs]
		, epochs = hp['epoch_count']
	)
	return model


def make_queue(repeat_count:int=1, fold_count:int=None):
	hyperparameters = {
		"include_2nd_dense": [True]
		, "neuron_multiply": [1.0]
		, "epoch_count": [10]
		, "learning_rate": [0.01]
		, "pool_size": [2]
		, "dropout": [0.4]
		, "batch_size": [8]
		, "kernel_size": [3]
		, "dense_neurons": [32]
		, "2nd_dense_neurons": [16]
		, "cnn_init": ['he_normal']
	}

	df = datum.to_pandas(name='brain_tumor.csv')
	# Dataset.Tabular
	dt_id = Dataset.Tabular.from_pandas(dataframe=df).id
	l_id = Label.from_dataset(dataset_id=dt_id, columns=['status']).id

	# Dataset.Image
	# Takes a while to run, but it tests both `from_urls` and `datum` functionality
	image_urls = datum.get_remote_urls(manifest_name='brain_tumor.csv')
	di_id = Dataset.Image.from_urls_pillow(urls=image_urls).id
	f_id = Feature.from_dataset(dataset_id=di_id).id
	
	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.18
		size_validation = 0.14

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = l_id
		, size_test = size_test
		, size_validation = size_validation
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(
			splitset_id=s_id, fold_count=fold_count
		).id
	else:
		fs_id = None

	a_id = Algorithm.make(
		library = "keras"
		, analysis_type = "classification_binary"
		, fn_build = fn_build
		, fn_train = fn_train
	).id

	h_id = Hyperparamset.from_algorithm(
		algorithm_id=a_id, hyperparameters=hyperparameters
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = h_id
		, repeat_count = repeat_count
	)
	return queue



"""PyTorch Binary Classification with Image data"""
# Internal modules
from .. import datum
from ..utils.pytorch import fit
from ..orm import *
# External modules
import torch
import torch.nn as nn
import torchmetrics


def fn_build(features_shape, label_shape, **hp):
	model = torch.nn.Sequential(
		#Conv1d(in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros')
		nn.Conv1d(
			in_channels=features_shape[0]#160 #running with `in_channels` as the width of the image. which is index[1], but only when batched?
			, out_channels=56 #arbitrary number. treating this as network complexity.
			, kernel_size=3
			, padding=1
		)
		, nn.ReLU() #it wasnt learning with tanh
		, nn.MaxPool1d(kernel_size=2, stride=2)
		, nn.Dropout(p=0.4)

		, nn.Conv1d(
			in_channels=56, out_channels=128,
			kernel_size=3, padding=1
		)
		, nn.ReLU()
		, nn.MaxPool1d(kernel_size=2, stride=2)
		, nn.Dropout(p=0.4)
		#[5x3840]
		, nn.Flatten()
		, nn.Linear(3840,3840)
		, nn.BatchNorm1d(3840,3840)
		, nn.ReLU()
		, nn.Dropout(p=0.4)

		, nn.Linear(3840, label_shape[0])
		, nn.Sigmoid()
	)
	return model


def fn_train(model, loser, optimizer, samples_train, samples_evaluate, **hp):   
	return fit(
		model, loser, optimizer, 
		samples_train, samples_evaluate,
		epochs=10, batch_size=5,
		metrics=[torchmetrics.Accuracy(),torchmetrics.F1Score()]
	)


def fn_predict(model, samples_predict):
	probability = model(samples_predict['features'])
	# Convert tensor back to numpy for AIQC metrics.
	probability = probability.detach().numpy()
	prediction = (probability > 0.5).astype("int32")
	# Both objects are numpy.
	return prediction, probability


def make_queue(repeat_count:int=1, fold_count:int=None):
	df = datum.to_pandas(name='brain_tumor.csv')
	# Dataset.Tabular
	dt_id = Dataset.Tabular.from_pandas(dataframe=df).id
	l_id = Label.from_dataset(dataset_id=dt_id, columns=['status']).id

	# Dataset.Image
	folder_path = 'remote_datum/image/brain_tumor/images'
	di_id = Dataset.Image.from_folder_pillow(
		folder_path=folder_path, ingest=False, dtype='float64'
	).id
	f_id = Feature.from_dataset(dataset_id=di_id).id
	FeatureShaper.from_feature(feature_id=f_id, reshape_indices=(0,2,3))
	
	if (fold_count is not None):
		size_test = 0.25
		size_validation = None
	elif (fold_count is None):
		size_test = 0.18
		size_validation = 0.14

	s_id = Splitset.make(
		feature_ids = [f_id]
		, label_id = l_id
		, size_test = size_test
		, size_validation = size_validation
	).id

	if (fold_count is not None):
		fs_id = Foldset.from_splitset(
			splitset_id=s_id, fold_count=fold_count
		).id
	else:
		fs_id = None

	a_id = Algorithm.make(
		library = "pytorch"
		, analysis_type = "classification_binary"
		, fn_build = fn_build
		, fn_train = fn_train
		, fn_predict = fn_predict
	).id

	queue = Queue.from_algorithm(
		algorithm_id = a_id
		, splitset_id = s_id
		, foldset_id = fs_id
		, hyperparamset_id = None #network takes a while.
		, repeat_count = repeat_count
	)
	return queue
